import logging
import os
import re
from collections.abc import Hashable
from typing import Any, List, Mapping, Optional, Union

import yaml
from yaml.constructor import ConstructorError

from configue.exceptions import NonCallableError, NotFoundError

# Matches "${myvar-default}" -> "${", "myvar", "-", "default", "}"
# or "${myvar}" -> "${", "myvar", "", "", "}"
ENV_PATTERN_REGEX = re.compile(r"(\${)(\w+)(-?)((?:(?![^}]*\${)[^}])*)(})")
CONSTRUCTOR_KEY = "()"
ESCAPED_CONSTRUCTOR_KEY = "\\()"


class ConfigueLoader(yaml.FullLoader):  # pylint: disable=too-many-ancestors
    logger = logging.getLogger(__name__)

    def construct_yaml_map(self, node: yaml.MappingNode) -> Any:
        mapping: Mapping[Hashable, Any] = self.construct_mapping(node)
        if isinstance(mapping, dict) and CONSTRUCTOR_KEY in mapping:
            path = mapping.pop(CONSTRUCTOR_KEY)
            object_path_elements = path.split(".")
            remaining_path_elements: List[str] = []
            exceptions: list[Optional[str]] = []
            while object_path_elements:
                try:
                    cls = self.find_python_name(
                        ".".join(object_path_elements),
                        node.start_mark,
                        unsafe=True,
                    )
                    break
                except ConstructorError as error:
                    exceptions.append(error.problem)
                    remaining_path_elements.insert(0, object_path_elements.pop(-1))
            else:
                raise NotFoundError(
                    f"Could not load element {path} {node.start_mark}, exceptions encountered: {exceptions!r}"
                ) from None
            for path_element in remaining_path_elements:
                cls = getattr(cls, path_element)

            if not callable(cls):
                raise NonCallableError(
                    f"Error while constructing a Python instance {node.start_mark}, "
                    f"expected a callable but found {type(cls)}"
                )
            return cls(**mapping)
        if isinstance(mapping, dict) and ESCAPED_CONSTRUCTOR_KEY in mapping:
            mapping[CONSTRUCTOR_KEY] = mapping.pop(ESCAPED_CONSTRUCTOR_KEY)
        return mapping

    def construct_scalar(self, node: Union[yaml.ScalarNode, yaml.MappingNode]) -> Any:
        scalar = yaml.FullLoader.construct_scalar(self, node)
        if isinstance(node, yaml.MappingNode):  # pragma: nocover
            return scalar
        replaced_value = ""
        end_pos = 0
        for match in ENV_PATTERN_REGEX.finditer(scalar):
            env_var_name, has_default, default_value = match.group(2, 3, 4)
            start_pos = match.start(1)
            if env_var_name not in os.environ and not has_default:
                self.logger.warning(f"Missing environment var: '{env_var_name}', no default is set")
            replaced_value += f"{scalar[end_pos:start_pos]}{os.environ.get(env_var_name, default_value)}"
            end_pos = match.end(5)
        replaced_value += scalar[end_pos:]
        if replaced_value == node.value:
            return scalar
        if node.style in ["'", '"']:
            replaced_value = f"{node.style}{replaced_value}{node.style}"
        # A variable has been replaced, reload to convert string to number if needed or replace again
        return yaml.load(replaced_value, Loader=self.__class__)



import logging
import os
from unittest import TestCase

import configue
from configue.exceptions import ConfigueError, NonCallableError, SubPathNotFound, NotFoundError
from tests.external_module import CONSTANT, MyObject, Color


class TestConfigue(TestCase):
    def tearDown(self) -> None:
        os.environ.pop("ENV_VAR", None)

    def test_load_file_shares_object_instances(self):
        result = configue.load(self._get_path("test_file_1.yml"), "key1")
        self.assertIs(result["subkey1"], result["subkey4"]["subkey5"])
        self.assertIsInstance(result["subkey1"], dict)
        self.assertIs(result["subkey1"], result["subkey4"]["subkey6"])
        self.assertIs(result["subkey2"], result["subkey4"]["subkey7"])
        self.assertIs(result["subkey2"], result["subkey4"]["subkey8"])
        self.assertIsInstance(result["subkey2"], MyObject)
        self.assertIs(result["subkey3"], result["subkey4"]["subkey9"])
        self.assertIs(result["subkey3"], result["subkey4"]["subkey10"])
        self.assertIsInstance(result["subkey3"], list)

    def test_load_deep_path(self):
        result = configue.load(self._get_path("test_file_1.yml"), "key1.subkey3.1")
        self.assertEqual("item2", result)

    def test_load_list_path(self):
        result = configue.load(self._get_path("test_file_1.yml"), ["key1", "sub.key.5"])
        self.assertEqual("final_value", result)

    def test_load_with_env_vars(self):
        result = configue.load(self._get_path("test_file_1.yml"), "env")
        self.assertEqual(
            {
                "env_key1": None,
                "env_key2": None,
                "env_key3": "",
                "env_key4": None,
                "env_key5": "default-value",
                "env_key6": 123,
                "env_key7": "123",
                "env_key8": "prepost",
                "env_key9": None,
                "env_key10": None,
            },
            result,
        )
        os.environ["ENV_VAR"] = "my_value"
        result = configue.load(self._get_path("test_file_1.yml"), "env")
        self.assertEqual(
            {
                "env_key1": "my_value",
                "env_key2": "my_value",
                "env_key3": "my_value",
                "env_key4": "my_value",
                "env_key5": "my_value",
                "env_key6": "my_value",
                "env_key7": "my_value",
                "env_key8": "premy_valuepost",
                "env_key9": "my_value",
                "env_key10": "my_value",
            },
            result,
        )
        os.environ["ENV_VAR"] = "321"
        result = configue.load(self._get_path("test_file_1.yml"), "env")
        self.assertEqual(
            {
                "env_key1": 321,
                "env_key2": 321,
                "env_key3": "321",
                "env_key4": 321,
                "env_key5": 321,
                "env_key6": 321,
                "env_key7": "321",
                "env_key8": "pre321post",
                "env_key9": 321,
                "env_key10": 321,
            },
            result,
        )
        os.environ["ENV_VAR"] = "${ENV_VAR_2}-${ENV_VAR_2}"
        os.environ["ENV_VAR_2"] = "123"
        result = configue.load(self._get_path("test_file_1.yml"), "env")
        self.assertEqual(
            {
                "env_key1": "123-123",
                "env_key2": "123-123",
                "env_key3": "123-123",
                "env_key4": "123-123",
                "env_key5": "123-123",
                "env_key6": "123-123",
                "env_key7": "123-123",
                "env_key8": "pre123-123post",
                "env_key9": "123-123",
                "env_key10": "123-123",
            },
            result,
        )

    def test_load_with_imports(self):
        os.environ["ENV_VAR"] = "test_file_1"
        result = configue.load(self._get_path("test_file_2.yml"), "key1")
        self.assertIs(result["value1"], result["value2"])
        self.assertIs(result["value1"], result["value3"])
        self.assertEqual("other_value", result["value1"]["key1"]["subkey1"]["other_key"])
        self.assertEqual("other_value", result["value4"]["other_key"])
        self.assertIsNone(result["value7"])

    def test_load_without_path(self):
        result = configue.load(self._get_path("test_file_1.yml"))
        self.assertCountEqual(["key1", "key2", "env"], result.keys())

    def test_load_with_invalid_class_raises_exception(self):
        with self.assertRaises(NonCallableError):
            configue.load(self._get_path("test_file_2.yml"), "invalid_class")

    def test_load_invalid_subpath_raises_exception(self):
        with self.assertRaises(SubPathNotFound):
            configue.load(self._get_path("test_file_1.yml"), "key1.subkey1.other_key.unknown_key")

        with self.assertRaises(SubPathNotFound):
            configue.load(self._get_path("test_file_1.yml"), "key1.subkey1.unknown_key")

        with self.assertRaises(SubPathNotFound):
            configue.load(self._get_path("test_file_1.yml"), "key1.subkey3.3")

        with self.assertRaises(SubPathNotFound):
            configue.load(self._get_path("test_file_1.yml"), "key1.subkey3.unknown_key")

    def test_load_invalid_import_raises_exception(self):
        with self.assertRaises(SubPathNotFound):
            configue.load(self._get_path("test_file_2.yml"), "invalid_import")

    def test_load_external_value(self):
        result = configue.load(self._get_path("test_file_2.yml"), "const")
        self.assertEqual(CONSTANT, result)

    def test_null_path(self):
        result = configue.load(self._get_path("test_file_2.yml"), "paths")
        self.assertIsNone(result["path"])
        self.assertEqual(os.path.expanduser("~"), result["path2"])
        self.assertIsNone(result["path3"])

    def test_logging_config(self):
        configue.load(self._get_path("test_file_2.yml"), "const", logging_config_path="logging_config")
        logger = logging.getLogger("test.path")
        self.assertEqual(logging.DEBUG, logger.handlers[0].level)
        self.assertEqual(logging.ERROR, logger.handlers[1].level)

    def test_load_internal_value_from_other_file(self):
        os.environ["ENV_VAR"] = "test_file_1"
        result = configue.load(self._get_path("test_file_2.yml"), "key1")
        self.assertEqual("other_value", result["value5"])
        self.assertEqual("other_value", result["value6"])

    def test_ext_enum_loading(self):
        result = configue.load(self._get_path("test_file_2.yml"), "enum_loading")
        self.assertEqual(Color.RED, result)

    def test_constructor_static_method_loading(self):
        result = configue.load(self._get_path("test_file_2.yml"), "static_loading")
        self.assertEqual("foo", result)

    def test_constructor_raises_exception_on_constructor_not_found(self):
        with self.assertRaises(NotFoundError):
            configue.load(self._get_path("test_file_2.yml"), "invalid_loading")

    def test_ext_raises_exception_on_module_not_found(self):
        with self.assertRaises(NotFoundError):
            configue.load(self._get_path("test_file_2.yml"), "invalid_ext.wrong_module")

    def test_ext_raises_exception_on_submodule_not_found(self):
        with self.assertRaises(NotFoundError):
            configue.load(self._get_path("test_file_2.yml"), "invalid_ext.wrong_sub_module")

    def test_ext_raises_exception_on_element_not_found(self):
        with self.assertRaises(ConfigueError):
            configue.load(self._get_path("test_file_2.yml"), "invalid_ext.wrong_element")

    def test_ext_raises_exception_on_property_not_found(self):
        with self.assertRaises(KeyError):
            configue.load(self._get_path("test_file_2.yml"), "invalid_ext.wrong_property")

    def test_loading_empty_file(self):
        content = configue.load(self._get_path("test_file_3.yml"))
        self.assertIsNone(content)

    @staticmethod
    def _get_path(file_name: str) -> str:
        return os.path.join(os.path.dirname(__file__), file_name)



import logging
import os
from typing import Any, List, Optional, TYPE_CHECKING, Type, Union, cast

from yaml import Loader, MappingNode, Node, ScalarNode, SequenceNode
from yaml.constructor import ConstructorError

from .configue_loader import ConfigueLoader
from .exceptions import SubPathNotFound, InvalidNodeType, NotFoundError

if TYPE_CHECKING:
    from .root_loader import RootLoader


class FileLoader:
    logger = logging.getLogger(__name__)

    def __init__(self, file_path: str, root_loader: "RootLoader") -> None:
        self._file_path = file_path
        self._root_loader = root_loader

        loader_cls: Type[Loader] = cast(
            Type[Loader],
            type("CustomLoader", (ConfigueLoader,), {"yaml_loader": self}),
        )

        loader_cls.add_multi_constructor("!import", self._load_import)  # type: ignore[no-untyped-call]
        loader_cls.add_constructor("!path", self._load_path)
        loader_cls.add_constructor("!cfg", self._load_cfg)
        loader_cls.add_constructor("!ext", self._load_ext)
        loader_cls.add_constructor("tag:yaml.org,2002:map", loader_cls.construct_yaml_map)  # type: ignore[type-var]

        with open(self._file_path, encoding="utf-8") as config_file:
            self._loader = loader_cls(config_file)
            self._root_node = self._loader.get_single_node()
        self._loader.dispose()  # type: ignore[no-untyped-call]

    def load(self, path: Union[str, List[str]]) -> Any:
        if self._root_node is None:
            return None

        if isinstance(path, str):
            path = path.split(".")

        current_node = self._root_node
        is_current_node_loaded = False
        for sub_path in path:
            if not sub_path:
                continue
            if not is_current_node_loaded:
                try:
                    current_node = self._get_node_at_sub_path(sub_path, current_node)
                except InvalidNodeType:
                    current_node = self._loader.construct_object(  # type: ignore[no-untyped-call]
                        current_node,
                        deep=True,
                    )
                    is_current_node_loaded = True
            if is_current_node_loaded:
                current_node = self._get_element_at_sub_path(sub_path, current_node)
        if is_current_node_loaded:
            return current_node
        return self._loader.construct_object(current_node, deep=True)  # type: ignore[no-untyped-call]

    @staticmethod
    def _get_node_at_sub_path(sub_path: str, current_node: Node) -> Node:
        if isinstance(current_node, SequenceNode):
            try:
                sub_path_index = int(sub_path)
            except ValueError:
                raise SubPathNotFound(
                    f"Could not convert sub_path element {sub_path} to list index {current_node.start_mark}"
                ) from None
            try:
                return cast(Node, current_node.value[sub_path_index])
            except IndexError:
                raise SubPathNotFound(
                    f"Could not find sub_path element {sub_path_index} in list {current_node.start_mark}"
                ) from None
        elif isinstance(current_node, MappingNode):
            for node_key, node_value in current_node.value:
                if isinstance(node_key, ScalarNode) and str(node_key.value) == sub_path:
                    return cast(Node, node_value)
            raise SubPathNotFound(f"Could not find sub_path {sub_path} {current_node.start_mark}")
        raise InvalidNodeType()

    @staticmethod
    def _get_element_at_sub_path(sub_path: str, current_element: Any) -> Any:
        try:
            sub_path_index: Union[str, int] = int(sub_path)
        except ValueError:
            sub_path_index = sub_path
        try:
            return current_element[sub_path_index]
        except TypeError:
            pass
        try:
            return getattr(current_element, sub_path)
        except AttributeError:
            raise SubPathNotFound(f"Could not find sub_path {sub_path} in {current_element}") from None

    def _load_import(self, loader: ConfigueLoader, tag_suffix: str, node: ScalarNode) -> Any:
        path = self._load_path(loader, node)
        if path is None:
            return None
        return self._root_loader.load_file(path, tag_suffix[1:])

    def _load_path(self, loader: ConfigueLoader, node: ScalarNode) -> Optional[str]:
        raw_path = loader.construct_scalar(node)
        if raw_path is None:
            return None
        path = os.path.expanduser(raw_path)
        return os.path.join(os.path.dirname(self._file_path), path)

    def _load_cfg(self, loader: ConfigueLoader, node: ScalarNode) -> Any:
        path = loader.construct_scalar(node)
        return self.load(path)

    def _load_ext(self, loader: ConfigueLoader, node: ScalarNode) -> Any:
        path = loader.construct_scalar(node)
        object_path_elements = path.split(".")
        remaining_path_elements: List[str] = []
        exceptions: list[Optional[str]] = []
        while object_path_elements:
            try:
                loaded_object = loader.find_python_name(
                    ".".join(object_path_elements),
                    node.start_mark,
                    unsafe=True,
                )
                break
            except ConstructorError as error:
                exceptions.append(error.problem)
                remaining_path_elements.insert(0, object_path_elements.pop(-1))
        else:
            raise NotFoundError(
                f"Could not load element {path} {node.start_mark}, encountered exceptions: {exceptions!r}"
            ) from None
        remaining_path = ".".join(remaining_path_elements)
        if remaining_path:
            return self._get_element_at_sub_path(remaining_path, loaded_object)
        return loaded_object
